/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *|

    pam_envfeed - set env vars in PAM based on an external program
    Copyright (C) 2004  Egmont Koblinger <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

|* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DEFAULT_PROG
#define DEFAULT_PROG "/sbin/pam_envfeed"
#endif

#ifndef BUFSIZE
#define BUFSIZE 4096
#endif

#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>
#include <stdarg.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#include <security/pam_modules.h>

extern char **environ;

static int nullterm = 0;
static int nouserinfo = 0;
static int debug = 0;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Are these really required, or just paranoia? */

static int x_open (const char *path, int flags)
{
	int i;
	do {
		i = open(path, flags);
	} while (i == -1 && errno == EINTR);
	return i;
}

static int x_close (int fd)
{
	int i;
	do {
		i = close(fd);
	} while (i == -1 && errno == EINTR);
	return i;
}

static int x_dup2 (int oldfd, int newfd)
{
	int i;
	do {
		i = dup2(oldfd, newfd);
	} while (i == -1 && errno == EINTR);
	return i;
}

static int x_pipe (int *filedes)
{
	int i;
	do {
		i = pipe(filedes);
	} while (i == -1 && errno == EINTR);
	return i;
}

static int x_fcntl (int fd, int cmd, long arg)
{
	int i;
	do {
		i = fcntl(fd, cmd, arg);
	} while (i == -1 && errno == EINTR);
	return i;
}

static ssize_t x_read (int fd, void *buf, size_t count)
{
	ssize_t i;
	do {
		i = read(fd, buf, count);
	} while (i == -1 && errno == EINTR);
	return i;
}

static ssize_t x_write (int fd, const void *buf, size_t count)
{
	ssize_t i;
	do {
		i = write(fd, buf, count);
	} while (i == -1 && errno == EINTR);
	return i;
}

static pid_t x_fork (void)
{
	pid_t i;
	do {
		i = fork();
	} while (i == -1 && errno == EINTR);
	return i;
}

static pid_t x_waitpid (pid_t pid, int *status, int options)
{
	pid_t i;
	do {
		i = waitpid(pid, status, options);
	} while (i == -1 && errno == EINTR);
	return i;
}

static int x_sigaction (int signum, const struct sigaction *act, struct sigaction *oldact)
{
	int i;
	do {
		i = sigaction(signum, act, oldact);
	} while (i == -1 && errno == EINTR);
	return i;
}

static int x_sigprocmask (int how, const sigset_t *set, sigset_t *oldset)
{
	int i;
	do {
		i = sigprocmask(how, set, oldset);
	} while (i == -1 && errno == EINTR);
	return i;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void log_msg (int err, const char *format, ...)
{
	va_list args;

	va_start(args, format);
	openlog("pam_envfeed", LOG_CONS | LOG_PID, LOG_AUTHPRIV);
	vsyslog(err, format, args);
	va_end(args);
	closelog();
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int pam_envfeed_child (int fd, int fdx,
                              const char *prog, char **env)
{
	int i, j;
	int reply[2];

	/* make sure fdx is at least 3 */
	if (fdx <= 2) {
		j = 3;
		if (fd == 3) j = 4;
		if (x_dup2(fdx, j) < 0) {
			reply[0] = 0;
			reply[1] = errno;
			goto err;
		}
		x_close(fdx);
		fdx = j;
	}

	/* we want fd to be stdout */
	if (fd != 1) {
		if (x_dup2(fd, 1) < 0) {
			reply[0] = 0;
			reply[1] = errno;
			goto err;
		}
	}

	/* clean up file descriptors */
	j = 1024;
	i = sysconf(_SC_OPEN_MAX);
	if (i < 0) {
		reply[0] = 1;
		reply[1] = 0;
		goto err;
	}
	if (i > j) j = i;
	for (i = 0; i < j; i++) {
		if (i != 1 && i != fdx) {
			x_close(i);
		}
	}
	if ((i = x_open("/dev/null", O_RDONLY)) != 0) {
		if (i < 0) {
			reply[0] = 2;
			reply[1] = errno;
		} else {
			reply[0] = 3;
			reply[1] = i;
		}
		goto err;
	}
	if ((i = x_open("/dev/null", O_WRONLY)) != 2) {
		if (i < 0) {
			reply[0] = 2;
			reply[1] = errno;
		} else {
			reply[0] = 3;
			reply[1] = i;
		}
		goto err;
	}

	/* c'mon */
	execle(prog, prog, NULL, env);
	reply[0] = 4;
	reply[1] = errno;

err:
	x_write(fdx, reply, sizeof(reply));
	return -1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int pam_envfeed (pam_handle_t *pamh, int flags, int argc,
                        const char **argv)
{
	const char *prog;
	int fd[2], fdx[2];
	pid_t pid;
	int status;
	int reply[2];
	int i, j;
	char *itemval;
	char **env;
	int realenvcount, pamenvcount, envcount;
	char buf[BUFSIZE];
	int buflen;
	char *buf2;
	int buf2len;
	sigset_t sigchld, oldset;
	struct sigaction sa;
	int sigchld_blocked, sigchld_caught;
	char * const *pam_environ;
	struct passwd pw;
	struct group gr;

	prog = DEFAULT_PROG;

	sigemptyset(&sigchld);
	sigaddset(&sigchld, SIGCHLD);
	sigchld_blocked = sigchld_caught = 0;
	pid = -1;
	fd[0] = fd[1] = fdx[0] = fdx[1] = -1;
	envcount = 0;
	env = NULL;
	buf2 = NULL;

	/* the boring part */
	for (i = 0; i < argc; i++) {
		if (!strncmp(argv[i], "prog=", 5)) {
			if (argv[i][5] != '/') {
				log_msg(LOG_CRIT, "Value of \"prog=\" has to be an absolute path");
				goto err;
			}
			prog = argv[i] + 5;
			continue;
		}
		if (!strcmp(argv[i], "nullterm")) {
			nullterm = 1;
			continue;
		}
		if (!strcmp(argv[i], "nouserinfo")) {
			nouserinfo = 1;
			continue;
		}
		if (!strcmp(argv[i], "debug")) {
			debug = 1;
			continue;
		}
		log_msg(LOG_CRIT, "Unknown option: %s", argv[i]);
		goto err;
	}

	realenvcount = 0;
	if (environ != NULL) {
		while (environ[realenvcount] != NULL) realenvcount++;
	}

	pamenvcount = 0;
	pam_environ = pam_getenvlist(pamh);
	if (pam_environ != NULL) {
		while (pam_environ[pamenvcount] != NULL) pamenvcount++;
	}

	envcount = realenvcount + pamenvcount + 11;

	if ((env = malloc((envcount + 1) * sizeof(char *))) == NULL) {
		log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
		goto err;
	}
	for (i = 0; i <= envcount; i++) env[i] = NULL;

	j = 0;
	for (i = 0; i < realenvcount; i++) {
		if ((env[j] = malloc(strlen(environ[i]) + 5)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "ENV_%s", environ[i]);
	}
	for (i = 0; i < pamenvcount; i++) {
		if ((env[j] = malloc(strlen(pam_environ[i]) + 8)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "PAMENV_%s", pam_environ[i]);
	}

	if (pam_get_item(pamh, PAM_RHOST, (const void **)&itemval) == PAM_SUCCESS && itemval != NULL) {
		if ((env[j] = malloc(strlen(itemval) + 11)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "PAM_RHOST=%s", itemval);
	}
	if (pam_get_item(pamh, PAM_RUSER, (const void **)&itemval) == PAM_SUCCESS && itemval != NULL) {
		if ((env[j] = malloc(strlen(itemval) + 11)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "PAM_RUSER=%s", itemval);
	}
	if (pam_get_item(pamh, PAM_SERVICE, (const void **)&itemval) == PAM_SUCCESS && itemval != NULL) {
		if ((env[j] = malloc(strlen(itemval) + 13)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "PAM_SERVICE=%s", itemval);
	}
	if (pam_get_item(pamh, PAM_TTY, (const void **)&itemval) == PAM_SUCCESS && itemval != NULL) {
		if ((env[j] = malloc(strlen(itemval) + 9)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "PAM_TTY=%s", itemval);
	}
	if (pam_get_item(pamh, PAM_USER, (const void **)&itemval) == PAM_SUCCESS && itemval != NULL) {
		if ((env[j] = malloc(strlen(itemval) + 10)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		sprintf(env[j++], "PAM_USER=%s", itemval);
	}
	if (!nouserinfo && itemval != NULL) {
		struct passwd *pwp = &pw;
		struct group *grp = &gr;
		buf2len = 1024;
		i = sysconf(_SC_GETPW_R_SIZE_MAX);
		if (i < 0) {
			log_msg(LOG_CRIT, "sysconf(_SC_GETPW_R_SIZE_MAX) not working");
			goto err;
		}
		if (i > buf2len) buf2len = i;
		i = sysconf(_SC_GETGR_R_SIZE_MAX);
		if (i < 0) {
			log_msg(LOG_CRIT, "sysconf(_SC_GETGR_R_SIZE_MAX) not working");
			goto err;
		}
		if (i > buf2len) buf2len = i;
		if ((buf2 = malloc(buf2len)) == NULL) {
			log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
			goto err;
		}
		getpwnam_r(itemval, pwp, buf2, buf2len, &pwp);
		if (pwp != NULL) {
			if (debug) {
				log_msg(LOG_DEBUG, "getpwnam was successful");
			}
			if (1) {
				if ((env[j] = malloc(24)) == NULL) {
					log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
					goto err;
				}
				sprintf(env[j++], "USERINFO_UID=%d", pw.pw_uid);
			}
			if (1) {
				if ((env[j] = malloc(24)) == NULL) {
					log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
					goto err;
					}
				sprintf(env[j++], "USERINFO_GID=%d", pw.pw_gid);
			}
			if (pw.pw_gecos != NULL) {
				if ((env[j] = malloc(strlen(pw.pw_gecos) + 16)) == NULL) {
					log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
					goto err;
				}
				sprintf(env[j++], "USERINFO_GECOS=%s", pw.pw_gecos);
			}
			if (pw.pw_dir != NULL) {
				if ((env[j] = malloc(strlen(pw.pw_dir) + 15)) == NULL) {
					log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
					goto err;
				}
				sprintf(env[j++], "USERINFO_HOME=%s", pw.pw_dir);
			}
			if (pw.pw_shell != NULL) {
				if ((env[j] = malloc(strlen(pw.pw_shell) + 16)) == NULL) {
					log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
					goto err;
				}
				sprintf(env[j++], "USERINFO_SHELL=%s", pw.pw_shell);
			}
			getgrgid_r(pw.pw_gid, grp, buf2, buf2len, &grp);
			if (grp != NULL) {
				if (debug) {
					log_msg(LOG_DEBUG, "getgrgid was successful");
				}
				if (gr.gr_name != NULL) {
					if ((env[j] = malloc(strlen(gr.gr_name) + 16)) == NULL) {
						log_msg(LOG_CRIT, "malloc failed: %s", strerror(errno));
						goto err;
					}
					sprintf(env[j++], "USERINFO_GROUP=%s", gr.gr_name);
				}
			} else {
				if (debug) {
					log_msg(LOG_DEBUG, "getgrgid didn't find the gid");
				}
			}
		} else {
			if (debug) {
				log_msg(LOG_DEBUG, "getpwnam didn't find the user");
			}
		}
	}

	/* block sigchld and make sure it isn't ignored */
	if (x_sigprocmask(SIG_BLOCK, &sigchld, &oldset) < 0) {
		log_msg(LOG_CRIT, "Failed to block signal: %s", strerror(errno));
		goto err;
	}
	if (!sigismember(&oldset, SIGCHLD)) {
		if (debug) {
			log_msg(LOG_DEBUG, "Blocking sigchld");
		}
		sigchld_blocked = 1;
	} else {
		if (debug) {
			log_msg(LOG_DEBUG, "Sigchld is already blocked");
		}
	}

	if (x_sigaction(SIGCHLD, NULL, &sa) < 0) {
		log_msg(LOG_CRIT, "Failed to get sigchld handler: %s", strerror(errno));
		goto err;
	}
	if (sa.sa_handler == SIG_IGN) {
		if (debug) {
			log_msg(LOG_DEBUG, "Sigchld was ignored, installing default handler");
		}
		memset(&sa, 0, sizeof(sa));
		sa.sa_handler = SIG_DFL;
		if (x_sigaction(SIGCHLD, &sa, NULL) < 0) {
			log_msg(LOG_CRIT, "Failed to set sigchld handler: %s", strerror(errno));
			goto err;
		}
		sigchld_caught = 1;
	} else {
		if (debug) {
			log_msg(LOG_DEBUG, "Sigchld was not ignored, handler left unchanged");
		}
	}

	/* prepare launching the child */
	if (debug) {
		log_msg(LOG_DEBUG, "About to launch %s", prog);
	}

	if (x_pipe(fd) < 0) {
		log_msg(LOG_CRIT, "Failed to create pipe: %s", strerror(errno));
		goto err;
	}
	if (x_pipe(fdx) < 0) {
		log_msg(LOG_CRIT, "Failed to create pipe: %s", strerror(errno));
		goto err;
	}
	if (x_fcntl(fdx[1], F_SETFD, FD_CLOEXEC) < 0) {
		log_msg(LOG_CRIT, "Failed to set close-on-exec bit: %s", strerror(errno));
		goto err;
	}

	/* fork */
	pid = x_fork();
	if (pid < 0) {
		log_msg(LOG_CRIT, "Failed to fork: %s", strerror(errno));
		goto err;
	}
	/* child */
	if (pid == 0) {
		x_close(fd[0]);
		x_close(fdx[0]);
		_exit(pam_envfeed_child(fd[1], fdx[1], prog, env));
	}
	/* parent goes on */
	x_close(fd[1]);
	fd[1] = -1;
	x_close(fdx[1]);
	fdx[1] = -1;

	/* see if child has successfully performed the exec */
	i = x_read(fdx[0], reply, sizeof(reply));
	if (i < 0) {
		log_msg(LOG_CRIT, "Error reading from pipe: %s", strerror(errno));
		goto err;
	}
	if (i == sizeof(reply)) {
		switch (reply[0]) {
		  case 0:
			log_msg(LOG_CRIT, "Child failed to perform dup2: %s", strerror(reply[1]));
			break;
		  case 1:
			log_msg(LOG_CRIT, "sysconf(_SC_OPEN_MAX) not working");
			break;
		  case 2:
			log_msg(LOG_CRIT, "Child failed to open /dev/null: %s", strerror(reply[1]));
			break;
		  case 3:
		  	log_msg(LOG_CRIT, "Child failed to open /dev/null");
		  	break;
		  case 4:
			log_msg(LOG_ALERT, "Child failed to exec %s: %s", prog, strerror(reply[1]));
			break;
		  default:
			log_msg(LOG_CRIT, "Internal error");
			break;
		}
		goto err;
	}
	if (i != 0) {
		log_msg(LOG_CRIT, "Internal error");
		goto err;
	}
	if (debug) {
		log_msg(LOG_DEBUG, "Child successfully exec'ed %s", prog);
	}
	x_close(fdx[0]);
	fdx[0] = -1;

	/* let's read the child's output */
	buflen = 0;
	while (1) {
		i = x_read(fd[0], buf + buflen, BUFSIZE - buflen);
		if (i < 0) {
			log_msg(LOG_CRIT, "Error reading from pipe: %s", strerror(errno));
			goto err;
		}
		if (i == 0) break;
		if (i == BUFSIZE - buflen) {
			log_msg(LOG_ALERT, "Child sent too long output");
			goto err;
		}
		buflen += i;
	}
	x_close(fd[0]);
	fd[0] = -1;

	/* wait for the child to terminate */
	if (debug) {
		log_msg(LOG_DEBUG, "Waiting for the child to terminate");
	}
	i = x_waitpid(pid, &status, 0);
	if (i < 0) {
		log_msg(LOG_CRIT, "Failed to waidpid: %s", strerror(errno));
		pid = -1;
		goto err;
	}
	if (i != pid) {
		log_msg(LOG_CRIT, "Failed to waidpid");
		pid = -1;
		goto err;
	}
	pid = -1;

	/* check the exit status */
	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) == 0) {
			if (debug) {
				log_msg(LOG_DEBUG, "Child returned with success");
			}
		} else {
			log_msg(LOG_ALERT, "Child returned exit code %d", WEXITSTATUS(status));
			goto err;
		}
	} else if (WIFSIGNALED(status)) {
		log_msg(LOG_ALERT, "Child exited on signal %d", WTERMSIG(status));
		goto err;
	} else {
		log_msg(LOG_ALERT, "Child exited abnormally");
		goto err;
	}

	/* restore sigchld handler and unblock it */
	if (sigchld_caught) {
		if (debug) {
			log_msg(LOG_DEBUG, "Ignoring sigchld");
		}
		sigchld_caught = 0;
		memset(&sa, 0, sizeof(sa));
		sa.sa_handler = SIG_IGN;
		if (x_sigaction(SIGCHLD, &sa, NULL) < 0) {
			log_msg(LOG_CRIT, "Failed to set sigchld handler: %s", strerror(errno));
			goto err;
		}
	}
	if (sigchld_blocked) {
		if (debug) {
			log_msg(LOG_DEBUG, "Unblocking sigchld");
		}
		sigchld_blocked = 0;
		if (x_sigprocmask(SIG_UNBLOCK, &sigchld, NULL) < 0) {
			log_msg(LOG_CRIT, "Failed to unblock sigchld: %s", strerror(errno));
			goto err;
		}
	}

	/* Everything was okay, set the environment variables */
	buf[buflen] = 0;
	if (!nullterm) {
		for (i = 0; i < buflen; i++) {
			if (buf[i] == '\n') buf[i] = 0;
		}
	}
	i = 0;
	while (1) {
		while (i < buflen && buf[i] == 0) i++;
		if (i == buflen) break;
		if (debug) {
			if (strchr(buf + i, '=') != NULL) {
				log_msg(LOG_DEBUG, "Setting env variable %s", buf + i);
			} else {
				log_msg(LOG_DEBUG, "Unsetting env variable %s", buf + i);
			}
		}
		pam_putenv(pamh, buf + i);
		i += strlen(buf + i);
	}


	for (i = 0; i < envcount; i++) {
		free(env[i]);
	}
	free(env);
	free(buf2);
	return 0;

err:
	if (pid > 0) {
		x_waitpid(pid, NULL, 0);
	}
	if (sigchld_caught) {
		memset(&sa, 0, sizeof(sa));
		sa.sa_handler = SIG_IGN;
		x_sigaction(SIGCHLD, &sa, NULL);
	}
	if (sigchld_blocked) {
		x_sigprocmask(SIG_UNBLOCK, &sigchld, NULL);
	}
	for (i = 0; i < envcount; i++) {
		free(env[i]);
	}
	free(env);
	free(buf2);
	if (fd[0] != -1) x_close(fd[0]);
	if (fd[1] != -1) x_close(fd[1]);
	if (fdx[0] != -1) x_close(fdx[0]);
	if (fdx[1] != -1) x_close(fdx[1]);
	return -1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* PAM_SM_AUTH */

PAM_EXTERN
int pam_sm_authenticate (pam_handle_t *pamh, int flags, int argc,
                         const char **argv)
{
	return PAM_IGNORE;
}

PAM_EXTERN
int pam_sm_setcred (pam_handle_t *pamh, int flags, int argc,
                    const char **argv)
{
	if (pam_envfeed(pamh, flags, argc, argv) < 0) {
		return PAM_CRED_ERR;
	}
	return PAM_SUCCESS;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* PAM_SM_ACCOUNT */

PAM_EXTERN
int pam_sm_acct_mgmt (pam_handle_t *pamh, int flags, int argc,
                      const char **argv)
{
	if (pam_envfeed(pamh, flags, argc, argv) < 0) {
		return PAM_AUTH_ERR;
	}
	return PAM_SUCCESS;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* PAM_SM_SESSION */

PAM_EXTERN
int pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
                         const char **argv)
{
	if (pam_envfeed(pamh, flags, argc, argv) < 0) {
		return PAM_SESSION_ERR;
	}
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_close_session (pam_handle_t *pamh, int flags, int argc,
                          const char **argv)
{
	return PAM_IGNORE;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Stuff for static linking */

#ifdef PAM_STATIC

struct pam_module _pam_envfeed_modstruct = {
	"pam_envfeed",
	pam_sm_authenticate,
	pam_sm_setcred,
	pam_sm_acct_mgmt,
	pam_sm_open_session,
	pam_sm_close_session,
	NULL
};

#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
